# -*- coding: utf-8 -*-
"""
Created on Mon Aug 10 10:23:56 2020

@author: Jean
"""

import numpy as np
import pandas as pd
from datetime import datetime
import matplotlib.pylab as plt
from statsmodels.tsa.stattools import adfuller
from statsmodels.tsa.seasonal import seasonal_decompose
from statsmodels.graphics.tsaplots import plot_acf, plot_pacf
from statsmodels.tsa.arima_model import ARIMA
import statsmodels.api as sm
import pmdarima as pm
import time

# 1 用pandas导入和处理时序数据 
# 读取数据，pd.read_csv默认生成DataFrame对象，需将其转换成Series对象
df = pd.read_csv("D:/ThinkPad_X1_Carbon/BaiduYunDownload/AirPassengers.csv", encoding='utf-8', index_col='Month')
# 将字符串索引转换成时间索引
df.index = pd.to_datetime(df.index)
# 生成pd.Series对象
ts = df['Passengers']  
ts.head()
# 查看某日的值既可以使用字符串作为索引，又可以直接使用时间对象作为索引
ts.index
ts['1949-01-01']
ts[datetime(1949,1,1)]
# 如果要查看某一年的数据，pandas也能非常方便的实现
ts['1949']
# 切片操作：
ts['1949-1' : '1949-6']
ts2 = ts.copy()
# 画时序图
ts2.plot(label="Passengers")
plt.title('AirPassengers Curve')
plt.legend(loc='best')
plt.show()
# 划分为测试集（1960）与训练集（1959以前）
test = ts['1960-1' : ]
ts  = ts[:'1959-12']
# 2 检验时序数据的稳定性
# 平稳性检验一般采用观察法和单位根检验法。
# A、观察法：需计算每个时间段内的平均的数据均值和标准差。
# 移动平均图
def draw_trend(timeseries, freq):
    f = plt.figure(facecolor='white')
    # 对size个数据进行移动平均
    rol_mean = timeseries.rolling(window=freq).mean()
    # 对size个数据移动平均的方差
    rol_std = timeseries.rolling(window=freq).std()
 
    timeseries.plot(color='blue', label='Original')
    rol_mean.plot(color='red', label='Rolling Mean')
    rol_std.plot(color='black', label='Rolling standard deviation')
    plt.legend(loc='best')
    plt.title('Rolling Mean & Standard Deviation')
    plt.show()
  
#Dickey-Fuller test:
def teststationarity(ts):
    dftest = adfuller(ts)
    # 对上述函数求得的值进行语义描述
    dfoutput = pd.Series(dftest[0:4], index=['Test Statistic','p-value',\
                        '#Lags Used','Number of Observations Used'])
    for key,value in dftest[4].items():
        dfoutput['Critical Value (%s)'%key] = value
    return dfoutput

draw_trend(ts,12)
# 通过上图，我们可以发现数据的移动平均值/标准差有越来越大的趋势，是不稳定的。
# B、接下来我们再看Dickey-Fuller的结果
teststationarity(ts)
# 此时p值为0.991880，说明并不能拒绝原假设。通过DF的数据可以明确的看出，在任何置信度下，数据都不是稳定的。

# 3 处理时序数据变成稳定数据
# 1）对数变换
ts_log = np.log(ts)
test_log = np.log(test)
# 划分为测试集（1960）与训练集（1959以前）
ts_log.plot(color='blue', label="log(Passengers)")
plt.title('log(Passengers) Curve')
plt.legend()
plt.show()
# 可以看出经过对数变换后，数据值域范围缩小了，振幅也没那么大了。

# 2）分解
def decompose(timeseries):
    # 返回包含三个部分 trend（趋势部分） ， seasonal（季节性部分） 和residual (残留部分)
    decomposition = seasonal_decompose(timeseries)
    trend = decomposition.trend
    seasonal = decomposition.seasonal
    residual = decomposition.resid
    
    plt.subplot(411)
    plt.plot(ts_log, label='Original')
    plt.legend(loc='best')
    plt.subplot(412)
    plt.plot(trend, label='Trend')
    plt.legend(loc='best')
    plt.subplot(413)
    plt.plot(seasonal,label='Seasonality')
    plt.legend(loc='best')
    plt.subplot(414)
    plt.plot(residual, label='Residuals')
    plt.legend(loc='best')
    plt.tight_layout()
    plt.show()
    return trend , seasonal, residual

# 分解成均值趋势、季节趋势及随机变动三个序列
trend , seasonal, residual = decompose(ts_log)
residual.dropna(inplace=True)
# residual的均值、方差
draw_trend(residual,12)
# residual的DF
teststationarity(residual)
# 如图所示，数据的均值和方差趋于常数，几乎无波动(看上去比之前的陡峭，但是要注意他的值域只有
# [-0.05,0.05]之间)，所以直观上可以认为是稳定的数据。另外DFtest的结果显示，
# Statistic值原小于1%时的Critical value，所以在99%的置信度下，数据是稳定的。

# 数据不稳定的原因主要有以下两点：
# A、趋势（trend）-数据随着时间变化。比如说升高或者降低。
# B、季节性(seasonality)-数据在特定的时间段内变动。比如说节假日，或者活动导致数据的异常。

# 3）平滑法
def draw_moving(timeSeries, size):
    f = plt.figure(facecolor='white')
    # 对size个数据进行移动平均
    rol_mean = timeSeries.rolling(window=size).mean()
    # 对size个数据进行加权移动平均
    # python 2.7
    # rol_weighted_mean = pd.ewma(timeSeries, span=size)
    # python 3.7
    rol_weighted_mean=timeSeries.ewm(halflife=size,min_periods=0,adjust=True,ignore_na=False).mean()
 
    timeSeries.plot(color='blue', label='Original')
    rol_mean.plot(color='red', label='Rolling Mean')
    rol_weighted_mean.plot(color='black', label='Weighted Rolling Mean')
    plt.legend(loc='best')
    plt.title('Rolling Mean')
    plt.show()
    
draw_moving(ts_log,12)
# 从上图可以发现窗口为12的移动平均能较好的剔除年周期性因素，
# 而指数平均法是对周期内的数据进行了加权，能在一定程度上减小年周期因素，
# 但并不能完全剔除，如要完全剔除可以进一步进行差分操作。

# 4）差分
# 剔除季节趋势
diff_12 = ts_log.diff(12)
diff_12.dropna(inplace=True)
# 剔除均值趋势
diff_12_1 = diff_12.diff(1)
diff_12_1.dropna(inplace=True)

draw_trend(diff_12_1,12)

teststationarity(diff_12_1)
# 从上面的统计检验结果可以看出，经过12阶差分和1阶差分后，该序列满足平稳性的要求了。

# 4 时序数据的预测


# 数据平稳后，需要对模型定阶，即确定p、q的阶数。先画出ACF,PACF的图像,代码如下：
def draw_acf_pacf(ts,lags):
    f = plt.figure(facecolor='white')
    ax1 = f.add_subplot(211)
    plot_acf(ts,ax=ax1,lags=lags)
    ax2 = f.add_subplot(212)
    plot_pacf(ts,ax=ax2,lags=lags)
    plt.subplots_adjust(hspace=0.5)
    plt.show()
    
draw_acf_pacf(diff_12_1,30)


# 观察上图，发现自相关和偏相系数都存在拖尾的特点，并且他们都具有明显的一阶相关性，所以我们设定p=1, q=1。
# 根据ts_diff_2的DF测试结果，对ts_diff_1进行d=1的1阶差分
# 建模与拟合ARIMA(0,1,0)ARIMA(1,1,1)^12模型
mod = sm.tsa.statespace.SARIMAX(ts_log, trend='n', order=(0,1,0), seasonal_order=(1,1,1,12))
results = mod.fit()
results.summary()
# 模型诊断，等均值，等方差，残差独立性通过
results.plot_diagnostics(lags=30, figsize=(16, 12))
# 残差稳定性检验
teststationarity(results.resid)

# 模型预测
# fitted values
ts1 = results.fittedvalues
log_recover_ts1 = np.exp(ts1)
# predict N steps into the future
steps = 12
start_index = len(ts1)
end_index = start_index + steps-1
forecasts = results.predict(start=start_index, end=end_index) 
# 对数还原
log_recover = np.exp(forecasts)
forecasts2 = test.copy()
forecasts2[:] = log_recover
# 画预测图
ts2.plot(color='blue',label="Original")
log_recover_ts1.plot(color='green',label="fitted")
forecasts2.plot(color='red',label="Prediction")
plt.title('AirPassengers ARIMA Prediction')    
plt.legend()
plt.show()



# Fit your model, 17.79s
# 建模
t1 = time.time()    
model = pm.auto_arima(ts_log, seasonal=True, m=12)
t2 = time.time()
print(round(t2-t1,2))
fitted = model.fit(ts_log)
# 参数显著性检验
fitted.summary()
# 模型诊断
fitted.plot_diagnostics(lags=30, figsize=(16, 12))
# 残差稳定性检验
teststationarity(fitted.arima_res_.resid)
# 模型预测
# fitted values
ts1 = ts_log.copy()
ts1[:] = fitted.arima_res_.fittedvalues
log_recover_ts1 = np.exp(ts1)
# predict N steps into the future
forecasts = model.predict(test_log.shape[0], return_conf_int=True) 
# 对数还原
log_recover = np.exp(forecasts[0])
forecasts2 = test.copy(); forecasts2[:] = log_recover
uper = np.exp(forecasts[1][:,1])
uper_recover = test.copy(); uper_recover[:] = uper
lower = np.exp(forecasts[1][:,0])
lower_recover = test.copy(); lower_recover[:] = lower
# 画预测图
ts2.plot(color='blue',label="Original")
log_recover_ts1.plot(color='green',label="fitted")
forecasts2.plot(color='red',label="Prediction")
plt.fill_between(forecasts2.index,uper_recover,lower_recover,alpha=0.9, color='orange',\
                     label="95% Confidence Intervals")
plt.title('AirPassengers ARIMA Prediction')    
plt.legend()
plt.show()

